/****************************************************************
\section{License}

This file is part of LIME.

LIME is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

LIME is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LIME.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/
/* \section{Overview} 
 *
 * This component is in all respects functionally equivalent
 * to the regular gain component. N.B. it does fix the bug 
 * that is present in the gain component.
 *
 * In this example however, we embed all documention as 
 * C comments to show that noweb literate 
 * programming system can function in the middle-end
 * rather than a back-end. To this end, the comments
 * should follow either of:
 *
 * \begin{enumerate}
 * \item use \TeX{} formatting (as this file does). This is probably the most
 *       usable for the UNIX project
 * \item use VERBATIM processing for cast (see LIME makefile for this). This
 *       only works well with stacks where people paid enough attention to
 *       comment quality...
 * \end{enumerate}
 *
 */
#include UNIX

#include <assert.h> /*
		       we use asserts here to check
		       the buffer availability at
		       run-time */

// the state of this processing element is defined elsewhere
struct gain_state;

// \ifx\HideCode\empty\else
static unsigned int Fsat(unsigned int,int); /* this static function
					     * implements saturated
					     * arithmetic */
// \fi

/* \section{Declaration}
 
   Here, actually, the original K\&R syntax makes a lot of
   sense, since the signature immediately specifies the ports,
   the processing element with its return type. The signature
   is not cluttered with ports' types and cardinalities, nevertheless,
   these are also readily visible in the text block that immediately
   follows the signature. */

void gain_process(buf,obuf,ins,outs)

/* \section{Specification}
 *
 * Now, the input/output ports' specifications follow, one per line 
 * to improve readability
 * \begin{enumerate}
 */

const unsigned int buf[restrict BUF_SIZE];	// \item input buffer
unsigned int obuf[restrict BUF_SIZE];		// \item output buffer
const struct gain_state ins[1];			// \item input state (note non-restrict)
struct gain_state outs[1];			// \item output state (note non-restrict)

// \end{enumerate}

// \ifx\HideCode\empty\else
// \section{Implementation}
//
// Implementation of the processing is split into two parts:
// \begin{enumerate}
// \item processing elements code, which iterates on the elements of the input buffer
// \item function that adjusts the value of an element using saturation arithmetic
// \end{enumerate}
//
// \subsection{Processing element}

{ // Here comes the processing element's code.

		/* Note how assert() calls are used to check the buffer availability,
or, alternatively to specify the activation rule for this component */

  assert(buf);
  assert(obuf);
  assert(ins);
  assert(outs);

/* Actual processing is here. */

/* Code is deliberately unindented to show that
 * the tooling can correct this. 
 */

{
unsigned int i;

for(i=0; i< BUF_SIZE; i++)
obuf[i]= Fsat(buf[i],ins[0].value);
}

outs[0].value=ins[0].value; /* pass the state to itself,
			       outs can overlap with ins,
			       so this is probably a no-op... */

 }

/* \subsection{Saturated arithmetic} */

#include <limits.h>
static unsigned int Fsat(unsigned int x,int y)
{
	if (y>0 && x>=UINT_MAX-y) return UINT_MAX;
	if (y<0 && x<=(unsigned)-y) return 0;
	return x+y;
}

// \fi
